from visad.python.JPythonMethods import *
import java.lang.Float as Float
import java.lang.Double as Double
import jarray, inspect
import traceback

#====================================================================================
def lineno():
    import inspect 

    """Returns the current line number in our program."""
    return inspect.currentframe().f_back.f_lineno
#====================================================================================


def rowColToIndex(row,column,numColumns):
   index = row*numColumns + column
   return index

def searchNeighbors(maskPixels, threshold=10):

  """ 
      clusterFlatField, numGroups, groupPoints = searchNeighbors(maskPixels, threshold=10)

      searchNeigbors:  accept pixel, find neighbors within a certain radius, 
                        find out those neighbors are in the list of badPixelss
                        group the pixels.

                        threshold:   Threshold of clustered pixels to filter.  If 
                                     below threshold, throw group in a noise category.
                        
                        maskPixels:  The array which contains the mask on input.
                                                                        
                                     
      Returns:          assignGrid:  The array which contains the grouping numbers.  
                                     NaN:  lat/lon nan, -1:  never considered for grouping, 
                                     -2: isolated,
                                     -5: below cluster threshold,
                                     1-n:  cluster groups
                                     
                        numGroups:   number of clusters found
                        groupPoints: the indices contained in each cluster group.  Returned
                                     as a list
                        nNoise:       number of clusters below threshold
                        noise:       indices of the noise
                        
     Written By:  Joleen Feltz (CIMSS/University of Wisconsin-Madison)
     Date:  March 24, 2013
     2D Clustering code adapted from CIMSS Cloud Object Tracking Code.
     Thank-you to Tom Whittaker and Tom Rink for helping improve code efficiency.

     Other Notes:
        There is a group 0, which should have no value.  This group is "skipped" so that no
        cluster group gets formally classified as 0.  Since masks and ADDE areas can contain
        zeros for values, it was not replicated in the assignment clusters to reduce possible
        confusion.  However, it does create confusion since jython arrays normally range from 0:N.      
  """
   
  print "SearchNeighbors Subroutine Begin" 

  radius=1

  # create the data array which will contain the groupings and adjust, if input is not a flatField.
  
  if (maskPixels.isFlatField()):
     newGrid = maskPixels.clone()
  else:
     newGrid = maskPixels[0].clone()

  # It is possible to multiply by -2.0 on an object, getFloats returns the clone of
  # an array.
  assignGrid = (-1.0 * newGrid).getFloats();

  # get the size of the flatField (numElements, numLines)
  [nx, ny] = getDomainSizes(newGrid)
  print "Domain and Range: ", nx,ny

  # get domain of this data set
  fieldDomain = newGrid.getDomainSet()

  # get the lat/lon values for the area
  fieldLatLonValues = GridUtil.getLatLon(fieldDomain)
   
 

  # just use the data, this a a tuple of [tupleDimension][x*y]
  assignGrid = assignGrid[0]
    
  useLine = []
  useElem = []
  group = 1
  # groupPoints becomes a list of dictionaries where the dictionary keys are the pixels and the values
  # is the group assigned to the pixel.  
  groupPoints = [{0:None}];
  groupPoints.append( {} );
  noise = [];
  nNoise = 0;
  areaOfGroup = {};
  groupUsed=False
  focusLine = 0;
  focusElem = 0;
  focusIndex = 0;
  incrementLine = 0;
  incrementElem = 0;
  checkLine = 0;
  checkElem = 0;
  checkIndex = 0;
 
  print  "In searchNeighbors"
  # loop over grid, starting one row/col in/down and ending before the last row/col

  for line in xrange(0,(ny-radius)+1):

    for elem in xrange(0,(nx-radius)+1):
      currentIndex = line*nx + elem;

      if (int(assignGrid[currentIndex]) == long(-1)):
       
        # Fix:  2013-05-09 Check for values off the earth
        if (Float.isNaN(fieldLatLonValues[0][currentIndex])) or (Float.isNaN(fieldLatLonValues[1][currentIndex])):
           assignGrid[currentIndex]=Double.NaN
        else:
           # assign the current grid point to the first value in the array to check for matches
           useLine.append(line)
           useElem.append(elem)
      
           while (len(useLine) > 0):  
             focusLine = useLine.pop(0)
             focusElem = useElem.pop(0)
             focusIndex = focusLine*nx + focusElem;

             for incrementLine in xrange(-radius,radius+1):
               for incrementElem in xrange(-radius, radius+1):
                 if (incrementElem == 0 and incrementLine == 0):
                   pass
   
                 else:
                   checkLine = focusLine + incrementLine
                   checkElem = focusElem + incrementElem
                   checkIndex = checkLine * nx + checkElem;
                   # check line/elem values to ensure they are within the array
                   if (checkLine >= 0 and checkElem >= 0 and checkLine < ny and checkElem < nx): 
                      # Fix:  2013-05-09 check this position for a mask value and valid earth location
                      if ((assignGrid[checkIndex] != 0.0) and not (Float.isNaN(fieldLatLonValues[0][checkIndex])) and not (Float.isNaN(fieldLatLonValues[1][checkIndex]))):
                        if (assignGrid[checkIndex] == -1): 
                          assignGrid[checkIndex] = group
                          assignGrid[focusIndex] = group
                          groupPoints[group][checkIndex] = 1;
                          groupPoints[group][focusIndex] = 1;
                          useLine.append(checkLine)
                          useElem.append(checkElem)
                          groupUsed = True
                          #print 'Line:  ' , lineno(), useLine, useElem, 'CurrentGroup=', group
                        elif (assignGrid[checkIndex] > 0):
                          assignGrid[focusIndex] = assignGrid[checkIndex]
                          groupPoints[group][focusIndex] = 1;
        
           if (groupUsed):
             #print "Group used = ",group;
             #print group, groupPoints[group].keys()
          
             # The keys of this dictionary are the actual indices for the group number.  
             #The dictionary is stored in a list indexed by group number.
             calcNPoints=len(groupPoints[group].keys());
             if (int(calcNPoints) >= threshold):
                # increment the group.  make space at the ending of the list for the next
                # group by adding a blank dictionary at the end, which will be filled with 
                # keys from the next group.
                group = group+1
                groupPoints.append( {} );
             else:
                nNoise = nNoise+1
                # pop this array from the list an forget about it?
                for forgetIndex in groupPoints[group].keys():
                   assignGrid[forgetIndex] = -5
                noise.append(groupPoints.pop(group))
                # do not increment the group, use this group number again.
                # make space for this group again.  add a blank dictionary at the 
                # end of the list
                groupPoints.append( {} );
             groupUsed=False

  #now write to a new data object (reshape the assignment grid into a 1D
  # array, assign the data samples in the newGrid (the flatField object cloned
  # earlier)  Do not make a copy, thus the trailing 0.
  
  newGrid.setSamples([assignGrid],0)
  print "SearchNeighbors Subroutine End"
  collectGarbage()
  
  return newGrid, group-1, groupPoints, nNoise, noise;
#**************************************************************************************
def reportCluster(inputMask, valueField, threshold=10, fileName="test.csv", fileDate=None):
   """ 
       assignGrid = reportCluster(inputMask, valueField, radius=1, threshold=10, fileName="test.csv")

       reportCluster:   Input mask to a a subroutine which finds cluster groups.  
                        Use cluster groups returned to report area of each cluster.  

                        Note:  This must be performed on a gridded flatfield.
                        
                        
                        inputMask:   The flatField which contains the MASKED (0/NaN or 0/1) data
                                     for clustering.

                        valueField:  The flatField which contains the data values.  (Usually, the
                                     original field used to create the mask - for example, if 
                                     a mask was created to screen for Aurora, this would be the
                                     Suomi NPP DNB radiances.)

                        radius:      Numer of pixels to search around current pixel 
                                     for clustering (input to searchNeighbors) 

                        threshold:   how many pixels must be in a group to get a group
                                     number assigned.

                        fileName:    Name for output CSV file which contains group
                                     number and area information.  (Default without a specific
                                     path will place a file in the McIDAS-V-System directory.

                        fileDate:    Optional java date string for reporting to file

      Returns:          assignGrid:  The array which contains the grouping numbers.  
                                     0: not in threshold 1:  within threshold, 
                                     NaN:  lat/lon nan, -2:  isolated
                                     2-n:  cluster groups
                        
     Written By:  Joleen Feltz (CIMSS/University of Wisconsin-Madison)
     Date:  March 24, 2013
     2D Clustering code adapted from clustering in CIMSS Cloud Object Tracking Code.
     Thank-you to Tom Whittaker helping improve code efficiency.
                        
   """

   import time
   import csv

   now = time.clock()
   print "starting at ",now
   emptyGroups = []

   # get domain of this data set.  

   if (inputMask.isFlatField()):
      fieldDomain = inputMask.getDomainSet()
   else:
      fieldDomain = inputMask[0].getDomainSet()

   if (valueField.isFlatField()):
      dataValues = valueField
   else:
      dataValues = valueField[0]

   # get the lat/lon values for the field (reported in csv later)
   latLons = GridUtil.getLatLon(fieldDomain)

   # create the Area Field from the flat field (to be used later when calculating area affected by bad pixels
   areaFlatField = createAreaField(inputMask)

   assignGrid, nGroups, groupIndexList, nNoise, noiseList = searchNeighbors(inputMask, threshold)

  
   # Create a new CSV file.
   csv.register_dialect("EscapeQuotes",quoting=csv.QUOTE_MINIMAL,doublequote=False,escapechar="\\")
   header1 = ['Group Number, TotalArea (m2), numPixels, mean, Lat/Lon[0], Lat/Lon[approxMidPoint], Lat/Lon[n-1]']
   
   try:
      stringDate = [fileDate.toString()]
   except:
      stringDate = str(type(fileDate))
   

   try:
      rowCount =  csv.reader(open(fileName))
      fileHandle = csv.writer(open(fileName, 'ab'),'EscapeQuotes')
      fileHandle.writerow(header1)
      fileHandle.writerow(stringDate)
   except IOError:
      fileHandle = csv.writer(open(fileName, 'wb'),'EscapeQuotes')
      fileHandle.writerow(header1)
      fileHandle.writerow(stringDate)


   
   sumNoise= 0;
   sumIndiv=0;
   sumAllGroups = 0;
   areaDict={}
   # see note in searchNeighbors, zero is skipped as an assignment value.
   for thisGroup in xrange(1, nGroups+1):
     
      areaDict[thisGroup] = computeSum(areaFlatField, groupIndexList[thisGroup].keys());

      # use the indices with the original data field to get a sum of data values for this group

      totalVals = computeSum(dataValues, groupIndexList[thisGroup].keys())

      #if (thisGroup == 1):
      #   print "1: ", len(groupIndexList[thisGroup].keys())
      #   print "2: ", len(sorted(groupIndexList[thisGroup]))

      if (areaDict[thisGroup] != 0.0):
         sortedKeys = sorted(groupIndexList[thisGroup])
         sizeThisGroup = len(sortedKeys)
         startPixel = sortedKeys[0]
         endPixel = sortedKeys[sizeThisGroup-1]
         if (sizeThisGroup >= 3):
            midN =  sortedKeys[sizeThisGroup/2 + 1]
            fileHandle.writerow([thisGroup, areaDict[thisGroup], sizeThisGroup, totalVals/sizeThisGroup, latLons[0][startPixel], latLons[1][startPixel],latLons[0][midN],latLons[1][midN], latLons[0][endPixel],latLons[1][endPixel]])
            sumAllGroups = sumAllGroups + areaDict[thisGroup]
         else:
          fileHandle.writerow([thisGroup, areaDict[thisGroup], sizeThisGroup, totalVals/sizeThisGroup, latLons[0][startPixel], latLons[1][startPixel], "Group < 3 points", "No Midpoint", latLons[0][endPixel],latLons[1][endPixel]])
      else:
         emptyGroups.append(thisGroup)
   

   indices = find(inputMask,'eq', 1)
   areaDict['total'] = computeSum(areaFlatField, indices)
   print "For all masked groups area = ",areaDict['total'];
   print "For groups above threshold = ",sumAllGroups

   indices2 = find(assignGrid, 'eq', -2);
   sumIndiv = computeSum(areaFlatField, indices2);
   indices2 = find(assignGrid, 'eq', -5);
   
   sumNoise = computeSum(areaFlatField, indices2);
   print "For individual pixels and clusters below threshold (aka:  Noise) = ", sumIndiv + sumNoise
   #print sumAllGroups+sumNoise+sumIndiv-areaDict['total']
  
   print "Group numbers with size of zero: ", emptyGroups, "Numbers groups below threshold of ",threshold, " pixels: ",nNoise
   print "Number of Groups: ", nGroups
   print "Done in ", (time.clock() - now), " seconds";

   return assignGrid